#ifndef __VESA__
#define __VESA__

#include <i86.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define FALSE	0
#define TRUE	1

void clear(int);

void *scrptr1;
void *scrptr2;
void *scrptr3;
void *scrptr;

#ifndef VESA_H
#define VESA_H

#pragma pack(1)

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

#define  __VBE_AVAILABLE   0x0001
#define  __VBE_TTYOUTPUT   0x0004
#define	__VBE_COLOR			0x0008
#define	__VBE_GRAPHICS		0x0010
#define	__VBE_NOTVGA		0x0020
#define  __VBE_NONBANKED   0x0040
#define	__VBE_LINEAR		0x0080

#define	__VBE_LINEAR_MODE	0x4000

typedef struct ___VBE_vesa_info_block {
	BYTE	VbeSignature[4];
	WORD	VbeVersion;
	DWORD	OemStringPtr;
	BYTE	Capabilities[4];
	DWORD	VideoModePtr;
	WORD	TotalMemory;
	WORD	OemSoftwareRev;
	DWORD	OemVendorNamePtr;
	DWORD	OemProductNamePtr;
	DWORD	OemProdcutRevPtr;
	BYTE	Reserved[222];
	BYTE	OemData[256];
} __VBE_vesa_info_block;

typedef struct ___VBE_mode_info_block {
	WORD	ModeAttributes;
	BYTE	WinAAttributes;
	BYTE	WinBAttributes;
	WORD	WinGranularity;
	WORD	WinSize;
	WORD	WinASegment;
	WORD	WinBSegment;
	DWORD	WinFuncPtr;
	WORD	BytesPerScanLine;

	WORD	XResolution;
	WORD	YResolution;
	BYTE	XCharSize;
	BYTE	YCharSize;
	BYTE	NumberOfPlanes;
	BYTE	BitsPerPixel;
	BYTE	NumberOfBanks;
	BYTE	MemoryModel;
	BYTE	BankSize;
	BYTE	NumberOfImagePages;
	BYTE	Reserved1;

	BYTE	RedMaskSize;
	BYTE	RedFieldPosition;
	BYTE	GreenMaskSize;
	BYTE	GreenFieldPosition;
	BYTE	BlueMaskSize;
	BYTE	BlueFieldPosition;
	BYTE	RsvdMaskSize;
	BYTE	RsvfFieldPosition;
	BYTE	DirectColorModeInfo;

	DWORD	PhysBasePtr;
	DWORD	OffScreenMemOffset;
	WORD	OffScreenMemSize;
	BYTE	Reserved2[206];
} __VBE_mode_info_block;

int __VBE_get_vesa_info( __VBE_vesa_info_block* );
int __VBE_get_mode_info( WORD , __VBE_mode_info_block* );
int __VBE_set_mode( WORD, void** );
int __VBE_set_display_start(int ,int);

#pragma pack()

#endif
#ifndef DPMI_H
#define DPMI_H

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

typedef struct ___dpmi_regs {
	DWORD	edi;
	DWORD	esi;
	DWORD	ebp;
	DWORD	res1;
	DWORD	ebx;
	DWORD	edx;
	DWORD	ecx;
	DWORD	eax;
	WORD	msw;
	WORD	es;
	WORD	ds;
	WORD	fs;
	WORD	gs;
	WORD	res2;
	WORD	res3;
	WORD	sp;
	WORD	ss;
} __dpmi_regs;

typedef struct ___dpmi_meminfo {
	DWORD	handle;
	DWORD	size;
	DWORD	address;
} __dpmi_meminfo;

int __dpmi_free_dos_memory( int );
int __dpmi_allocate_dos_memory( int , int* );
void __dpmi_simulate_real_mode_interrupt( int , __dpmi_regs* );
int __dpmi_physical_address_mapping( __dpmi_meminfo* );

#endif


static int RMseg = 0;
static int PMsel = 0;
void __VBE_shut_down( void );

int __VBE_get_vesa_info( __VBE_vesa_info_block* info ){
	__dpmi_regs r;
	memset( &r, 0, sizeof( __dpmi_regs ) );

    if ( PMsel == 0 ){
		RMseg = __dpmi_allocate_dos_memory( (1024+15)>>4, &PMsel );
		fprintf( stderr, "DOS MEM: %x:0000\n", RMseg );
        atexit( __VBE_shut_down );
    }

	info->VbeSignature[0] = 'V';
	info->VbeSignature[1] = 'B';
	info->VbeSignature[2] = 'E';
	info->VbeSignature[3] = '2';

	// Call the VESA Info Function 
	r.eax = 0x4F00;
	r.edi = 0;
	r.es = RMseg;

    _fmemcpy(MK_FP(PMsel,0),info,sizeof(__VBE_vesa_info_block));
	__dpmi_simulate_real_mode_interrupt( 0x10, &r );
    _fmemcpy(info,MK_FP(PMsel,0),sizeof(__VBE_vesa_info_block));

	return TRUE;
}

int __VBE_set_display_start(int x_pos,int y_pos){

	__dpmi_regs r;
	memset( &r, 0, sizeof( __dpmi_regs ) );

	// Call the VESA Info Function 
    r.eax = 0x4F07;  //get/set display-start
    r.ebx = 0x0080;  //set display-start during vertical retrace
    r.ecx = x_pos;
    r.edx = y_pos;

	__dpmi_simulate_real_mode_interrupt( 0x10, &r );

	return TRUE;
}

int __VBE_get_mode_info( WORD mode, __VBE_mode_info_block* info ){
	__dpmi_regs r;
	memset( &r, 0, sizeof( __dpmi_regs ) );

    if ( PMsel == 0 ){
		RMseg = __dpmi_allocate_dos_memory( (1024+15)>>4, &PMsel );
        atexit( __VBE_shut_down );
    }

	// Call the VESA Info Function 
	r.eax = 0x4F01;
	r.ecx = mode;
	r.edi = 0;
	r.es = RMseg;

    _fmemcpy(MK_FP(PMsel,0),info,sizeof(__VBE_mode_info_block));
	__dpmi_simulate_real_mode_interrupt( 0x10, &r );
    _fmemcpy(info,MK_FP(PMsel,0),sizeof(__VBE_mode_info_block));

	return TRUE;
}

int __VBE_set_mode( WORD mode, void** addr ){
	__dpmi_regs r;

	memset( &r, 0, sizeof( __dpmi_regs ) );

	r.eax = 0x4F02;

    if ( addr ){
		r.ebx = mode | __VBE_LINEAR_MODE;
    }
    else{
		r.ebx = mode;
    }

	__dpmi_simulate_real_mode_interrupt( 0x10, &r );

    if ( addr ){
		__dpmi_meminfo mapInfo;
		__VBE_mode_info_block info;

        if ( !__VBE_get_mode_info( mode, &info ) ){
            return FALSE;
        }

		mapInfo.size = 1024 * 1024 * 2;
		mapInfo.address = info.PhysBasePtr;

		__dpmi_physical_address_mapping( &mapInfo );
		*addr = (void*) mapInfo.address;
    }

	return TRUE;
}

void __VBE_shut_down( void ){
	__dpmi_free_dos_memory( PMsel );
}

int __dpmi_free_dos_memory( int PMsel )
	{
    union REGS  r;
    r.w.ax = 0x101;
    r.w.dx = PMsel;
    int386(0x31, &r, &r);
	return 0;
	}

int __dpmi_allocate_dos_memory( int para, int* PMsel )
	{
    union REGS r;
    r.w.ax = 0x100;
    r.w.bx = para;
    int386(0x31, &r, &r);
    if (r.w.cflag)
		{
	    *PMsel = 0;
	    return -1;
		}

    *PMsel = r.w.dx;
    return r.w.ax;
	}

void __dpmi_simulate_real_mode_interrupt( int no, __dpmi_regs* regs )
	{
	union REGS reg;
	struct SREGS sreg;

//    segread(&sreg);
	memset( &sreg, 0, sizeof( struct SREGS ) );
    reg.w.ax = 0x300;
    reg.h.bl = no;
    reg.h.bh = 0;
    reg.w.cx = 0;
	sreg.es = FP_SEG( regs );
//	sreg.es = sreg.ds;
  reg.x.edi = FP_OFF( regs );
//	reg.x.edi = regs;

    int386x(0x31, &reg, &reg, &sreg);
	}

int __dpmi_physical_address_mapping( __dpmi_meminfo* info )
	{
    union REGS  r;

	// Setup mapping
    r.w.ax = 0x800;
    r.w.bx = info->address >> 16;
    r.w.cx = info->address & 0xFFFF;
    r.w.si = info->size >> 16;
    r.w.di = info->size & 0xFFFF;
    int386(0x31, &r, &r);

    if (r.x.cflag)
		{
		fprintf( stderr, "__dpmi_physical_address_mapping failed!\n" );
		return -1;
		}
	
   	info->address = ((long)r.w.bx << 16) + r.w.cx;
	return 0;
	}

void set640x480x16(){
  __VBE_vesa_info_block vInfo;
  __VBE_mode_info_block mInfo;
  char* addr = NULL;

  if ( !__VBE_get_vesa_info( &vInfo ) ){
    fprintf(stderr,"Error retrieving vesa info\n" );
    exit(1);
  }

  if((vInfo.TotalMemory*64*1024)<(640*480*2*3)){
      printf("need %d Mb of video memory!\n",640*480*2*3);
    exit(1);
  }

  if ( !__VBE_get_mode_info( 0x111 , &mInfo ) ){
    fprintf(stderr,"Error retrieving vesa info\n" );
    exit(1);
  }

  if ( !__VBE_set_mode( 0x111, (void**) &addr ) ){
    fprintf(stderr,"Error setting vesa info\n" );
    exit(1);
  }
		   
  scrptr1=((short*)addr)+640*480*0;
  scrptr2=((short*)addr)+640*480*1;
  scrptr3=((short*)addr)+640*480*2;
  scrptr=scrptr1;
}

void set640x480x8(){
  __VBE_vesa_info_block vInfo;
  __VBE_mode_info_block mInfo;
  char* addr = NULL;

  if ( !__VBE_get_vesa_info( &vInfo ) ){
    fprintf(stderr,"Error retrieving vesa info\n" );
    exit(1);
  }

  if((vInfo.TotalMemory*64*1024)<(640*480*3)){
      printf("need %d Mb of video memory!\n",640*480*3);
    exit(1);
  }

  if ( !__VBE_get_mode_info( 0x101, &mInfo ) ){
    fprintf(stderr,"Error retrieving vesa info\n" );
    exit(1);
  }

  if ( !__VBE_set_mode( 0x101, (void**) &addr ) ){
    fprintf(stderr,"Error setting vesa info\n" );
    exit(1);
  }
		   
  scrptr1=((unsigned char*)addr)+640*480*0;
  scrptr2=((unsigned char*)addr)+640*480*1;
  scrptr3=((unsigned char*)addr)+640*480*2;
  scrptr=scrptr1;
}

void set320x200x24(){
  __VBE_vesa_info_block vInfo;
  __VBE_mode_info_block mInfo;
  char* addr = NULL;

  if ( !__VBE_get_vesa_info( &vInfo ) ){
    fprintf(stderr,"Error retrieving vesa info\n" );
    exit(1);
  }

  if((vInfo.TotalMemory*64*1024)<(320*200*3*3)){
      printf("need %d Mb of video memory!\n",320*200*3*3);
    exit(1);
  }

  if ( !__VBE_get_mode_info( 0x10F, &mInfo ) ){
    fprintf(stderr,"Error retrieving vesa info\n" );
    exit(1);
  }

  if ( !__VBE_set_mode( 0x10F, (void**) &addr ) ){
    fprintf(stderr,"Error setting vesa info\n" );
    exit(1);
  }
		   
  scrptr1=((unsigned char*)addr)+320*200*0;
  scrptr2=((unsigned char*)addr)+320*200*1*3;
  scrptr3=((unsigned char*)addr)+320*200*2*3;
  scrptr=scrptr1;
}

void set320x200x16(){
  __VBE_vesa_info_block vInfo;
  __VBE_mode_info_block mInfo;
  char* addr = NULL;

  if ( !__VBE_get_vesa_info( &vInfo ) ){
    fprintf(stderr,"Error retrieving vesa info\n" );
    exit(1);
  }

  if((vInfo.TotalMemory*64*1024)<(320*200*2*3)){
      printf("need %d Mb of video memory!\n",320*200*2*3);
    exit(1);
  }

  if ( !__VBE_get_mode_info( 0x10E, &mInfo ) ){
    fprintf(stderr,"Error retrieving vesa info\n" );
    exit(1);
  }

  if ( !__VBE_set_mode( 0x10E, (void**) &addr ) ){
    fprintf(stderr,"Error setting vesa info\n" );
    exit(1);
  }
		   
  scrptr1=((short*)addr)+320*200*0;
  scrptr2=((short*)addr)+320*200*1;
  scrptr3=((short*)addr)+320*200*2;
  scrptr=scrptr1;
}


void pageFlip(){
    if(scrptr==scrptr1){
        __VBE_set_display_start(0,480*0);
        scrptr=scrptr2;
    }
    else if(scrptr==scrptr2){
        __VBE_set_display_start(0,480*1);
        scrptr=scrptr3;
    }
    else if(scrptr==scrptr3){
        __VBE_set_display_start(0,480*2);
        scrptr=scrptr1;
    }
    return;
}

void pageFlip320x200(){
    if(scrptr==scrptr1){
        __VBE_set_display_start(0,200*0);
        scrptr=scrptr2;
    }
    else if(scrptr==scrptr2){
        __VBE_set_display_start(0,200*1);
        scrptr=scrptr3;
    }
    else if(scrptr==scrptr3){
        __VBE_set_display_start(0,200*2);
        scrptr=scrptr1;
    }
    return;
}

#endif
